#include <iostream>
#include <sstream>
#include <string>

#include <SDL/SDL.h>
#include <GL/gl.h>
#include <GL/glext.h>

#include "Shader.h"
#include "ShaderProgram.h"
#include "FrameTimer.h"
#include "VertexBuffer.h"
#include "FrameBuffer.h"
#include "Texture.h"
#include "StringTex.h"
#include "HdrFile.h"

using std::cout;
using std::endl;
using std::stringstream;
using std::istringstream;
using std::string;


int main(int argc, char **argv)
{
	SDL_Surface *screen;
	SDL_Event event;
	Uint8 *keystate;
	const SDL_VideoInfo* info = NULL;
	int width; // Size determined by
	int height; // radiance texture
	int bpp = 0;
	int flags = 0;

	int quit = 0;
	stringstream ss;
	string str;
	bool show_text = true;

	const float default_exposure = 1.0f;
	const float exposure_step = 1.0f;
	const float exposure_smallstep = 0.1f;
	float exposure = default_exposure;
	const float default_factor = 0.4f;
	const float factor_step = 0.1f;
	float factor = default_factor;
	const float default_gamma = 0.55f;
	const float gamma_step = 0.1f;
	float gamma = default_gamma;

	FrameTimer t(5, 100);

	HdrFile hdr_file;

// ----- Parsing shell argument ---------------
	if (argc != 2) {
		cout << "Provide exactly an argument!" << endl;
		exit(-1);
	}
	else {
		hdr_file.Load(argv[1]);
		width = hdr_file.Width();
		height = hdr_file.Height();
	}


// ----- SDL init ---------------
	if (SDL_Init(SDL_INIT_VIDEO) < 0) {
		fprintf(stderr, "Video initialization failed: %s\n", SDL_GetError());
		exit(-1);
	}

	atexit(SDL_Quit);
	
	info = SDL_GetVideoInfo();
	bpp = info->vfmt->BitsPerPixel;

	
// ----- OpenGL attribute setting via SDL ---------------
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	SDL_GL_SetAttribute(SDL_GL_SWAP_CONTROL, 0);

	flags = SDL_OPENGL;


// ----- Setting up the screen surface ---------------
	if ((screen = SDL_SetVideoMode(width, height, bpp, flags)) == 0) {
		fprintf(stderr, "Video mode set failed: %s\n", SDL_GetError());
        	exit(-1);
	}

	SDL_WM_SetCaption("HDR_load1 by Encelo", NULL);


// ----- OpenGL init ---------------
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glDisable(GL_DEPTH_TEST);
	glBlendFunc(GL_ONE, GL_ONE);

	ShaderProgram simple_sh("shaders/simple.vert", "shaders/simple.frag");
	ShaderProgram hipass_sh("shaders/simple.vert", "shaders/hdr1/hipass.frag");
	ShaderProgram gaussH_sh("shaders/simple.vert", "shaders/gaussianH.frag");
	ShaderProgram gaussV_sh("shaders/simple.vert", "shaders/gaussianV.frag");
	ShaderProgram tonemap_sh("shaders/simple.vert", "shaders/hdr1/tonemap.frag");


	float weights[9] = {0.0677841f, 0.0954044f, 0.121786f, 0.140999f, 0.148054f, 0.140999f, 0.121786f, 0.0954044f, 0.0677841f};
	float offsetsH[9], offsetsV[9];
	for(int i=0; i<9; i++) {
		offsetsH[i] = (i - 4.0f) / float(width/2.0f);
		offsetsV[i] = (i - 4.0f) / float(height/2.0f);
	}
	gaussH_sh.Uniform("Weights", 9, weights);
	gaussH_sh.Uniform("Offsets", 9, offsetsH);
	gaussV_sh.Uniform("Weights", 9, weights);
	gaussV_sh.Uniform("Offsets", 9, offsetsV);

	tonemap_sh.Uniform("Tex0", 0);
	tonemap_sh.Uniform("Tex1", 1);


	StringTex string_tex("fonts/blaklitn.ttf", 24);

	Texture hdr_tex(GL_RGB16F_ARB, hdr_file.Width(), hdr_file.Height(), hdr_file.Image());

	FrameBuffer *fbo = new FrameBuffer[4];
	fbo[0].SetSize(width, height);
	fbo[0].AttachRender(GL_DEPTH_COMPONENT24);
    fbo[0].AttachTexture(GL_RGBA16F_ARB);
	for (int i = 1; i < 4; i++) {
		fbo[i].SetSize(int(width/2), int(height/2));
		fbo[i].AttachTexture(GL_RGBA16F_ARB, GL_LINEAR);
		// clamping to preserve post-pro from blur leaking
		fbo[i].BindTex();
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

	}
	FrameBuffer::Unbind();


	VertexBuffer quad;
	GLfloat qverts[] = {-1,-1, 0,  1,-1, 0,  1, 1, 0,  -1, 1, 0};
	GLfloat qcoords[] = {0, 0,  1, 0,  1, 1,  0, 1};
	GLuint qind[] = {0, 1, 2,  0, 2, 3};
	quad.SetVertices(4, qverts);
	quad.SetTexCoords(4, qcoords);
	quad.SetIndices(6, qind);

	t.Reset();


// ----- Event cycle ---------------
	while (!quit) {
		while (SDL_PollEvent(&event)) {
			keystate = SDL_GetKeyState(NULL);
			switch (event.type) {
			case SDL_QUIT:
				quit = true;
				break;
			case SDLK_SPACE:
				show_text = !show_text;
				break;
			case SDL_KEYDOWN:
				switch (event.key.keysym.sym) {
				case SDLK_ESCAPE:
				case SDLK_q:
					quit = true;
					break;
				case SDLK_F1:
					SDL_WM_ToggleFullScreen(screen);
					break;
				case SDLK_SPACE:
					show_text = !show_text;
					break;
				case SDLK_KP7:
					if (keystate[SDLK_LSHIFT] && exposure >= exposure_smallstep) {
						exposure -= exposure_smallstep;
					}
					else if (exposure >= exposure_step) {
						exposure -= exposure_step;
					}
					break;
				case SDLK_KP8:
					exposure = default_exposure;
					break;
				case SDLK_KP9:
					if (keystate[SDLK_LSHIFT]) {
						exposure += exposure_smallstep;;
					}
					else {
						exposure += exposure_step;;
					}
					break;
				case SDLK_KP4:
					if (factor >= factor_step) {
						factor -= factor_step;
					}
					break;
				case SDLK_KP5:
					factor = default_factor;
					break;
				case SDLK_KP6:
					if (factor < 1.0f) {
						factor += factor_step;
					}
					break;
				case SDLK_KP1:
					if (gamma >= gamma_step) {
						gamma -= gamma_step;
					}
					break;
				case SDLK_KP2:
					gamma = default_gamma;
					break;
				case SDLK_KP3:
					gamma += gamma_step;
					break;
				case SDLK_KP0:
					exposure = default_exposure;
					factor = default_factor;
					gamma = default_gamma;
					break;
				default:
					break;
				}
			}
		}


// ----- Blitting on the screen ---------------
		t.AddFrame();

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();

		// First Pass: scene rendering
		fbo[0].BindOutput();
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		simple_sh.Use();
		glActiveTexture(GL_TEXTURE0);
		hdr_tex.Bind();
		glMatrixMode(GL_TEXTURE);
		glScalef(1.0f, -1.0f, 1.0f);
		quad.DrawElements();
		glScalef(1.0f, -1.0f, 1.0f);
		glMatrixMode(GL_MODELVIEW);

		// Second Pass: downsampling and hipass filter
		fbo[1].BindOutput();
		glClear(GL_COLOR_BUFFER_BIT);
		fbo[0].BindTex();
		hipass_sh.Use();
		glViewport(0, 0, width/2, height/2);
		quad.DrawElements();

		// Third Pass: gaussian blur along the X axis
		fbo[2].BindOutput();
		glClear(GL_COLOR_BUFFER_BIT);
		fbo[1].BindTex();
		gaussH_sh.Use();		
		quad.DrawElements();

		// Fourth Pass: gaussian blur along the Y axis
		fbo[3].BindOutput();
		glClear(GL_COLOR_BUFFER_BIT);
		fbo[2].BindTex();
		gaussV_sh.Use();		
		quad.DrawElements();

		// Fifth Pass: tone mapping and composition
		FrameBuffer::Unbind();
		glClear(GL_COLOR_BUFFER_BIT);
		glActiveTexture(GL_TEXTURE0);
		fbo[0].BindTex();
		glActiveTexture(GL_TEXTURE1);
		fbo[3].BindTex();
		tonemap_sh.Use();
		tonemap_sh.Uniform("Exposure", exposure);
		tonemap_sh.Uniform("Factor", factor);
		tonemap_sh.Uniform("Gamma", gamma);
		glViewport(0, 0, width, height);
		quad.DrawElements();


		// Overlay text
		glActiveTexture(GL_TEXTURE0);
		glClear(GL_DEPTH_BUFFER_BIT);
		glEnable(GL_BLEND);
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glOrtho(0, width, height, 0, -1, 1);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glTranslatef(10.0f, float(height-120), 0.0f);
		simple_sh.Use();

		ss << exposure;
		str = "Exposure: " + ss.str() + "\n";
		ss.clear(istringstream::goodbit);
		ss.str("");
	
		ss << factor;		
		str += "Factor: " + ss.str() + "\n";
		ss.clear(istringstream::goodbit);
		ss.str("");

		ss << gamma;
		str += "Gamma: " + ss.str() + "\n";
		ss.clear(istringstream::goodbit);
		ss.str("");

		ss << t.GetFPS();
		str += "FPS: " + ss.str() ;
		ss.clear(istringstream::goodbit);
		ss.str("");

		if (show_text)
			string_tex.Draw(str);

		glDisable(GL_BLEND);

		SDL_GL_SwapBuffers();
	}


// ----- Quitting ---------------
	delete[] fbo;
	SDL_Quit();
	return 0;	
}
