#include <iostream>
#include <sstream>
#include <string>

#include <SDL/SDL.h>
#include <GL/gl.h>
#include <GL/glext.h>

#include "Shader.h"
#include "ShaderProgram.h"
#include "FrameTimer.h"
#include "VertexBuffer.h"
#include "FrameBuffer.h"
#include "Texture.h"
#include "StringTex.h"
#include "HdrFile.h"

using std::cout;
using std::endl;
using std::stringstream;
using std::istringstream;
using std::string;


int main(int argc, char **argv)
{
	SDL_Surface *screen;
	SDL_Event event;
	Uint8 *keystate;
	const SDL_VideoInfo* info = NULL;
	int width;  // Size determined by
	int height; // radiance texture
	int bpp = 0;
	int flags = 0;

	int quit = 0;
	stringstream ss;
	string str;
	bool show_text = true;

	const float default_exposure = 1.0f;
	const float exposure_step = 1.0f;
	const float exposure_smallstep = 0.1f;
	float exposure = default_exposure;
	const float default_threshold = 0.5f;
	const float threshold_step = 0.1f;
	float threshold = default_threshold;

	FrameTimer t(5, 100);

	HdrFile hdr_file;

// ----- Parsing shell argument ---------------
	if (argc != 2) {
		cout << "Provide exactly an argument!" << endl;
		exit(-1);
	}
	else {
		hdr_file.Load(argv[1]);
		width = hdr_file.Width();
		height = hdr_file.Height();
	}


// ----- SDL init ---------------
	if (SDL_Init(SDL_INIT_VIDEO) < 0) {
		fprintf(stderr, "Video initialization failed: %s\n", SDL_GetError());
		exit(-1);
	}

	atexit(SDL_Quit);
	
	info = SDL_GetVideoInfo();
	bpp = info->vfmt->BitsPerPixel;

	
// ----- OpenGL attribute setting via SDL ---------------
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	SDL_GL_SetAttribute(SDL_GL_SWAP_CONTROL, 0);

	flags = SDL_OPENGL;


// ----- Setting up the screen surface ---------------
	if ((screen = SDL_SetVideoMode(width, height, bpp, flags)) == 0) {
		fprintf(stderr, "Video mode set failed: %s\n", SDL_GetError());
        	exit(-1);
	}

	SDL_WM_SetCaption("HDR_load2 by Encelo", NULL);


// ----- OpenGL init ---------------
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glDisable(GL_DEPTH_TEST);
	glBlendFunc(GL_ONE, GL_ONE);

	ShaderProgram simple_sh("shaders/simple.vert", "shaders/simple.frag");
	ShaderProgram luminance2_sh("shaders/simple.vert", "shaders/hdr2/luminance2x2.frag");
	ShaderProgram luminance3_sh("shaders/simple.vert", "shaders/hdr2/luminance3x3.frag");
	ShaderProgram bright_sh("shaders/simple.vert", "shaders/hdr2/brightpass.frag");
	ShaderProgram gaussH_sh("shaders/simple.vert", "shaders/gaussianH.frag");
	ShaderProgram gaussV_sh("shaders/simple.vert", "shaders/gaussianV.frag");
	ShaderProgram tonemap_sh("shaders/simple.vert", "shaders/hdr2/tonemap.frag");
	ShaderProgram addrgb_sh("shaders/simple.vert", "shaders/hdr2/addrgb.frag");


	luminance2_sh.Uniform("Width", width);
	luminance2_sh.Uniform("Height", height);
	luminance3_sh.Uniform("Width", width);
	luminance3_sh.Uniform("Height", height);
	bright_sh.Uniform("Tex0", 0);
	bright_sh.Uniform("Tex1", 1);

	float weights[9] = {0.0677841f, 0.0954044f, 0.121786f, 0.140999f, 0.148054f, 0.140999f, 0.121786f, 0.0954044f, 0.0677841f};
	float offsetsH[9], offsetsV[9];
	for (int i=0; i<9; i++) {
		offsetsH[i] = (i - 4.0f) / float(width);
		offsetsV[i] = (i - 4.0f) / float(height);
	}

	gaussH_sh.Uniform("Weights", 9, weights);
	gaussH_sh.Uniform("Offsets", 9, offsetsH);
	gaussV_sh.Uniform("Weights", 9, weights);
	gaussV_sh.Uniform("Offsets", 9, offsetsV);

	tonemap_sh.Uniform("Tex0", 0);
	tonemap_sh.Uniform("Tex1", 1);
	addrgb_sh.Uniform("Tex0", 0);
	addrgb_sh.Uniform("Tex1", 1);


	StringTex string_tex("fonts/blaklitn.ttf", 24);
	Texture hdr_tex(GL_RGB16F_ARB, hdr_file.Width(), hdr_file.Height(), hdr_file.Image());


	// Create full-size scene, bright filtered, and tone-mapped FBOs
	FrameBuffer scene_fbo(width, height);
	scene_fbo.AttachRender(GL_DEPTH_COMPONENT24);
	scene_fbo.AttachTexture(GL_RGBA16F_ARB);

	FrameBuffer bright_fbo(width, height);
	bright_fbo.AttachTexture(GL_RGBA16F_ARB);
	FrameBuffer tone_fbo(width, height);
	tone_fbo.AttachTexture(GL_RGBA16F_ARB);

	// Create full-size blur FBOs
	FrameBuffer *blur_fbo = new FrameBuffer[2];
	for (int i = 0; i < 2; i++) {
		blur_fbo[i].SetSize(width, height);
		blur_fbo[i].AttachTexture(GL_RGBA16F_ARB);
		// clamping to preserve post-pro from blur leaking
		blur_fbo[i].BindTex();
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	}

	// Create luminance FBOs chain
	int chain_length = 1;
	int start_size = (height/4 < width/4) ? height/4 : width/4; // min(width/4, height/4)
	int size = 1;
	for (size = 1; size < start_size; size *= 3) {
		chain_length++;
	}
	size /= 3;
	chain_length--;

	int max3_size = size;

	FrameBuffer lum2_fbo(width/4, height/4);
	lum2_fbo.AttachTexture(GL_RGBA16F_ARB, GL_NEAREST);

	FrameBuffer *lum3_fbo = new FrameBuffer[chain_length];
	for (int i = 0; i < chain_length; i++) {
		lum3_fbo[i].SetSize(size, size);
		lum3_fbo[i].AttachTexture(GL_RGBA16F_ARB, GL_NEAREST);
		size /= 3;
	}

	FrameBuffer::Unbind();


	// Screen aligned quad VBO creation and init
	VertexBuffer quad;
	GLfloat qverts[] = {-1,-1, 0,  1,-1, 0,  1, 1, 0,  -1, 1, 0};
	GLfloat qcoords[] = {0, 0,  1, 0,  1, 1,  0, 1};
	GLuint qind[] = {0, 1, 2,  0, 2, 3};
	quad.SetVertices(4, qverts);
	quad.SetTexCoords(4, qcoords);
	quad.SetIndices(6, qind);

	t.Reset();


// ----- Event cycle ---------------
	while (!quit) {
		while (SDL_PollEvent(&event)) {
			keystate = SDL_GetKeyState(NULL);
			switch (event.type) {
			case SDL_QUIT:
				quit = true;
				break;
			case SDLK_SPACE:
				show_text = !show_text;
				break;
			case SDL_KEYDOWN:
				switch (event.key.keysym.sym) {
				case SDLK_ESCAPE:
				case SDLK_q:
					quit = true;
					break;
				case SDLK_F1:
					SDL_WM_ToggleFullScreen(screen);
					break;
				case SDLK_SPACE:
					show_text = !show_text;
					break;
				case SDLK_KP7:
					if (keystate[SDLK_LSHIFT] && exposure >= exposure_smallstep) {
						exposure -= exposure_smallstep;
					}
					else if (exposure >= exposure_step) {
						exposure -= exposure_step;
					}
					break;
				case SDLK_KP8:
					exposure = default_exposure;
					break;
				case SDLK_KP9:
					if (keystate[SDLK_LSHIFT]) {
						exposure += exposure_smallstep;
					}
					else {
						exposure += exposure_step;
					}
					break;
				case SDLK_KP4:
					if (threshold >= threshold_step) {
						threshold -= threshold_step;
					}
					break;
				case SDLK_KP5:
					threshold = default_threshold;
					break;
				case SDLK_KP6:
					if (threshold < 2.9f) {
						threshold += threshold_step;
					}
					break;
				case SDLK_KP0:
					exposure = default_exposure;
					threshold = default_threshold;
					break;
				default:
					break;
				}
			}
		}


// ----- Blitting on the screen ---------------
		t.AddFrame();

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();

		// First Pass: scene rendering
		scene_fbo.BindOutput();
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		simple_sh.Use();
		glActiveTexture(GL_TEXTURE0);
		hdr_tex.Bind();
		glMatrixMode(GL_TEXTURE);
		glScalef(1.0f,-1.0f, 1.0f);
		quad.DrawElements();
		glScalef(1.0f,-1.0f, 1.0f);
		glMatrixMode(GL_MODELVIEW);

		// Second Pass: calculating average and maximum scene luminance
		luminance2_sh.Use();
		lum2_fbo.BindOutput();
		glClear(GL_COLOR_BUFFER_BIT);
		scene_fbo.BindTex();
		glViewport(0, 0, width/4, height/4);
		quad.DrawElements();

		size = max3_size;

		luminance3_sh.Use();
		lum2_fbo.BindTex();
		luminance3_sh.Uniform("Width", width/4);
		luminance3_sh.Uniform("Height", height/4);
		for (int i = 0; i < chain_length; i++) {
			lum3_fbo[i].BindOutput();
			glClear(GL_COLOR_BUFFER_BIT);
			glViewport(0, 0, size, size);
			quad.DrawElements();

			luminance3_sh.Uniform("Width", size);
			luminance3_sh.Uniform("Height", size);
			lum3_fbo[i].BindTex();
			size /= 3;
		}

		// Third Pass: bright pass filter
		bright_sh.Use();
		bright_sh.Uniform("Exposure", exposure);
		bright_sh.Uniform("Threshold", threshold);
		glActiveTexture(GL_TEXTURE0);
		scene_fbo.BindTex();
		glActiveTexture(GL_TEXTURE1);
		lum3_fbo[chain_length-1].BindTex(); // Single pixel luminance FBO
		bright_fbo.BindOutput();
		glClear(GL_COLOR_BUFFER_BIT);
		glViewport(0, 0, width, height);
		quad.DrawElements();

		// Fourth Pass: gaussian blur along the X axis
		gaussH_sh.Use();
		blur_fbo[0].BindOutput();
		glClear(GL_COLOR_BUFFER_BIT);
		glActiveTexture(GL_TEXTURE0);
		bright_fbo.BindTex();
		quad.DrawElements();

		// Fifth Pass: gaussian blur along the Y axis
		gaussV_sh.Use();
		blur_fbo[1].BindOutput();
		glClear(GL_COLOR_BUFFER_BIT);
		blur_fbo[0].BindTex();
		quad.DrawElements();

		// Sixth Pass: tone mapping
		tonemap_sh.Use();
		tonemap_sh.Uniform("Exposure", exposure);
		tone_fbo.BindOutput();
		glClear(GL_COLOR_BUFFER_BIT);
		glActiveTexture(GL_TEXTURE0);
		scene_fbo.BindTex();
		glActiveTexture(GL_TEXTURE1);
		lum3_fbo[chain_length-1].BindTex(); // Single pixel luminance FBO
		glViewport(0, 0, width, height);
		quad.DrawElements();

		// Seventh Pass: glow effect
		addrgb_sh.Use();
		FrameBuffer::Unbind();
		glClear(GL_COLOR_BUFFER_BIT);
		glActiveTexture(GL_TEXTURE0);
		tone_fbo.BindTex();
		glActiveTexture(GL_TEXTURE1);
		blur_fbo[1].BindTex();
		quad.DrawElements();


		// Overlay text
		glActiveTexture(GL_TEXTURE0);
		glClear(GL_DEPTH_BUFFER_BIT);
		glEnable(GL_BLEND);
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glOrtho(0, width, height, 0, -1, 1);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glTranslatef(10.0f, float(height-100), 0.0f);
		simple_sh.Use();

		ss << exposure;
		str = "Exposure: " + ss.str() + "\n";
		ss.clear(istringstream::goodbit);
		ss.str("");
	
		ss << threshold;		
		str += "Threshold: " + ss.str() + "\n";
		ss.clear(istringstream::goodbit);
		ss.str("");

		ss << t.GetFPS();
		str += "FPS: " + ss.str() ;
		ss.clear(istringstream::goodbit);
		ss.str("");

		if (show_text)
			string_tex.Draw(str);

		glDisable(GL_BLEND);

		SDL_GL_SwapBuffers();
	}


// ----- Quitting ---------------
	delete[] lum3_fbo;
	delete[] blur_fbo;
	SDL_Quit();
	return 0;	
}
