#include <string>
#include <iostream>
#include <stdexcept>

#include <SDL/SDL.h>
#include <SDL/SDL_ttf.h>
#include <GL/gl.h>

#include "StringTex.h"

using std::string;
using std::vector;
using std::invalid_argument;
using std::cout;
using std::endl;

#include "Timer.h" // XXX: for debug purposes only

StringTex::StringTex(const string& filename, int size) throw(invalid_argument): 
	_vertices(NULL), _texcoords(NULL), _indices(NULL)
{
	Timer t; // XXX: for debug purposes only
	int i;
	char ch[2];
	TTF_Font *font;
	SDL_Surface *temp, *image;
	SDL_Color white = {255, 255, 255};

	const int num_chars = 95;
	_glyphs = new Glyph[num_chars];

	_vertices = new GLfloat[3 * 4 * num_chars];
	_texcoords = new GLfloat[2 * 4 * num_chars];
	_indices = new GLuint[6 * num_chars];

	TTF_Init();
	std::cout << "Loading " << filename << std::endl;
	font = TTF_OpenFont(filename.c_str(), size);
	if(!font) {
		throw invalid_argument("Text::Text - Cannot open the font");
	}

	_line_skip = TTF_FontLineSkip(font);


	t.Start();
	ch[1] = '\0';
	for(i = 0; i < num_chars; i++) {
		ch[0] = i + 32; // 32 because ' ' is a character

		temp = TTF_RenderText_Blended(font, ch, white);
		_glyphs[i].w = temp->w;
		_glyphs[i].h = temp->h;

		SetupVBOChar(i); // call after setting glyph data

		image = SDL_CreateRGBSurface(0, temp->w, temp->h, 32, 0x00ff0000, 0x0000ff00, 0x000000ff, 0xff000000);
//		SDL_FillRect(image, NULL, (2*i+5)*256*256 + 256*(4*i+3) + i); // texture background to help in debugging
//		cout << "char: " << ch << "  temp_w = " << temp->w << "  temp_h = " << temp->h << endl;
		SDL_BlitSurface(temp, NULL, image, NULL);
		SDL_FreeSurface(temp);

		_tex.push_back(new Texture(image));
		SDL_FreeSurface(image);
	}
	t.Stop();
	cout << "Character textures creation: " << t.GetInterval() << "ms" << endl;

	TTF_CloseFont(font);
	TTF_Quit(); //XXX: sdl_ttf init and quit at each object creation

	_vbo.SetVertices(4 * num_chars, _vertices);
	_vbo.SetTexCoords(4 * num_chars, _texcoords);
	_vbo.SetIndices(6 * num_chars, _indices);
	delete[] _indices;
	delete[] _texcoords;
	delete[] _vertices;
	_vbo.SetBinding(false);
}


StringTex::~StringTex()
{
	vector<Texture *>::const_iterator cii;
	for(cii = _tex.begin(); cii != _tex.end(); cii++) {
		delete *cii;
	}

	delete[] _glyphs;
}


void StringTex::SetupVBOChar(int idx)
{

	// 12 vertices components per quad
	_vertices[idx*12 + 0] = _glyphs[idx].w; // x
	_vertices[idx*12 + 1] = _glyphs[idx].h; // y
	_vertices[idx*12 + 2] = 0;              //_z_
	_vertices[idx*12 + 3] = 0;              // x
	_vertices[idx*12 + 4] = _glyphs[idx].h; // y
	_vertices[idx*12 + 5] = 0;              //_z_
	_vertices[idx*12 + 6] = 0;              // x
	_vertices[idx*12 + 7] = 0;              // y
	_vertices[idx*12 + 8] = 0;              //_z_
	_vertices[idx*12 + 9] = _glyphs[idx].w; // x
	_vertices[idx*12 + 10] = 0;             // y
	_vertices[idx*12 + 11] = 0;             //_z_

	// 8 texcoords components per quad
	_texcoords[idx*8 + 0] = 1;              // s
	_texcoords[idx*8 + 1] = 1;              //_t_
	_texcoords[idx*8 + 2] = 0;              // s
	_texcoords[idx*8 + 3] = 1;              //_t_
	_texcoords[idx*8 + 4] = 0;              // s
	_texcoords[idx*8 + 5] = 0;              //_t_
	_texcoords[idx*8 + 6] = 1;              // s
	_texcoords[idx*8 + 7] = 0;              //_t_

	// 6 indices per quad
	_indices[idx*6 + 0] = idx*4 + 0;
	_indices[idx*6 + 1] = idx*4 + 1;
	_indices[idx*6 + 2] = idx*4 + 2;        //___
	_indices[idx*6 + 3] = idx*4 + 0;
	_indices[idx*6 + 4] = idx*4 + 2;
	_indices[idx*6 + 5] = idx*4 + 3;
}


void StringTex::Draw(const string& text)
{
	int w;
	int linew = 0; // needed to go back on a new line
	int index;

	_vbo.Bind();

	glPushMatrix();
	for(int i=0; i < int(text.length()); i++) {
		if (text[i] == '\n') {
			glTranslatef(-linew, _line_skip, 0.0f);
			linew = 0;
			continue;
		}

		index = int(text[i]) - 32;
		w = _glyphs[index].w;

		_tex[index]->Bind();
		_vbo.DrawElements(index * 6, 6);

		linew += w;
		glTranslatef(w, 0.0f, 0.0f);
	}
	glPopMatrix();
}
